# Copyright 2024- Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base classes that need to be available everywhere.

Mainly tooling around configuration, protocols and superclasses for References
and WorkflowComponents, that are concretized elsewhere.
"""

from dataclasses import dataclass
import importlib
import base64
from attrs import define
from functools import lru_cache
from typing import (
    Generic,
    TypeVar,
    Protocol,
    Iterator,
    Unpack,
    TypedDict,
    NotRequired,
    Generator,
    Any,
    get_args,
    get_origin,
    Annotated,
    Callable,
    cast,
    runtime_checkable,
)
from contextlib import contextmanager
from contextvars import ContextVar
from sympy import Expr, Symbol, Basic
import copy

BasicType = str | float | bool | bytes | int | None
RawType = BasicType | list["RawType"] | dict[str, "RawType"]
FirmType = RawType | list["FirmType"] | dict[str, "FirmType"] | tuple["FirmType", ...]
# Basic is from Sympy, which does not have type annotations, so ExprType cannot pass mypy
ExprType = (FirmType | Basic | list["ExprType"] | dict[str, "ExprType"] | tuple["ExprType", ...])  # type: ignore # fmt: skip

U = TypeVar("U")
T = TypeVar("T")


def strip_annotations(parent_type: type) -> tuple[type, tuple[str]]:
    """Discovers and removes annotations from a parent type.

    Args:
        parent_type: a type, possibly Annotated.

    Returns: a parent type that is not Annotation, along with any stripped metadata, if present.

    """
    # Strip out any annotations. This should be auto-flattened, so in theory only one iteration could occur.
    metadata = []
    while get_origin(parent_type) is Annotated:
        parent_type, *parent_metadata = get_args(parent_type)
        metadata += list(parent_metadata)
    return parent_type, tuple(metadata)


# Generic type for configuration settings for the renderer
RenderConfiguration = dict[str, RawType]


class WorkflowProtocol(Protocol):
    """Expected structure for a workflow.

    We do not expect various workflow implementations, but this allows us to define the
    interface expected by the core classes.
    """

    def remap(self, name: str) -> str:
        """Perform any name-changing for steps, etc. in the workflow.

        This enables, for example, simplifying all the IDs to an integer sequence.

        Returns: remapped name.
        """
        ...

    def set_result(self, result: Basic | list[Basic] | tuple[Basic]) -> None:
        """Set the step that should produce a result for the overall workflow."""
        ...

    def simplify_ids(self, infix: list[str] | None = None) -> None:
        """Drop the non-human-readable IDs if possible, in favour of integer sequences.

        Args:
            infix: any inherited intermediary identifiers, to allow nesting, or None.
        """
        ...


class BaseRenderModule(Protocol):
    """Common routines for all renderer modules."""

    @staticmethod
    def default_config() -> dict[str, RawType]:
        """Retrieve default settings.

        These will not change during execution, but can be overridden by `dewret.core.set_render_configuration`.

        Returns: a static, serializable dict.
        """
        ...


@runtime_checkable
class RawRenderModule(BaseRenderModule, Protocol):
    """Render module that returns raw text."""

    def render_raw(
        self, workflow: WorkflowProtocol, **kwargs: RenderConfiguration
    ) -> dict[str, str]:
        """Turn a workflow into flat strings.

        Returns: one or more subworkflows with a `__root__` key representing the outermost workflow, at least.
        """
        ...


@runtime_checkable
class StructuredRenderModule(BaseRenderModule, Protocol):
    """Render module that returns JSON/YAML-serializable structures."""

    def render(
        self, workflow: WorkflowProtocol, **kwargs: RenderConfiguration
    ) -> dict[str, dict[str, RawType]]:
        """Turn a workflow into a serializable structure.

        Returns: one or more subworkflows with a `__root__` key representing the outermost workflow, at least.
        """
        ...


class RenderCall(Protocol):
    """Callable that will render out workflow(s)."""

    def __call__(
        self, workflow: WorkflowProtocol, **kwargs: RenderConfiguration
    ) -> dict[str, str] | dict[str, RawType]:
        """Take a workflow and turn it into a set of serializable (sub)workflows.

        Args:
            workflow: root workflow.
            kwargs: configuration for the renderer.

        Returns: a mapping of keys to serialized workflows, containing at least `__root__`.
        """
        ...


class UnevaluatableError(Exception):
    """Signposts that a user has tried to treat a reference as the real (runtime) value.

    For example, by comparing to a concrete integer or value, etc.
    """

    ...


@define
class ConstructConfiguration:
    """Basic configuration of the construction process.

    Holds configuration that may be relevant to `construst(...)` calls or, realistically,
    anything prior to rendering. It should hold generic configuration that is renderer-independent.
    """

    flatten_all_nested: bool = False
    allow_positional_args: bool = False
    allow_plain_dict_fields: bool = False
    field_separator: str = "/"
    field_index_types: str = "int"
    simplify_ids: bool = False


class ConstructConfigurationTypedDict(TypedDict):
    """Basic configuration of the construction process.

    Holds configuration that may be relevant to `construst(...)` calls or, realistically,
    anything prior to rendering. It should hold generic configuration that is renderer-independent.

    **THIS MUST BE KEPT IDENTICAL TO ConstructConfiguration.**
    """

    flatten_all_nested: NotRequired[bool]
    allow_positional_args: NotRequired[bool]
    allow_plain_dict_fields: NotRequired[bool]
    field_separator: NotRequired[str]
    field_index_types: NotRequired[str]
    simplify_ids: NotRequired[bool]


@define
class GlobalConfiguration:
    """Overall configuration structure.

    Having a single configuration dict allows us to manage only one ContextVar.
    """

    construct: ConstructConfiguration
    render: dict[str, RawType]


CONFIGURATION: ContextVar[GlobalConfiguration] = ContextVar("configuration")


@contextmanager
def set_configuration(
    **kwargs: Unpack[ConstructConfigurationTypedDict],
) -> Iterator[ContextVar[GlobalConfiguration]]:
    """Sets the construct-time configuration.

    This is a context manager, so that a setting can be temporarily overridden and automatically restored.
    """
    with _set_configuration() as CONFIGURATION:
        for key, value in kwargs.items():
            setattr(CONFIGURATION.get().construct, key, value)
        yield CONFIGURATION


@contextmanager
def set_render_configuration(
    kwargs: dict[str, RawType],
) -> Iterator[ContextVar[GlobalConfiguration]]:
    """Sets the render-time configuration.

    This is a context manager, so that a setting can be temporarily overridden and automatically restored.

    Returns: the yielded global configuration ContextVar.
    """
    with _set_configuration() as CONFIGURATION:
        CONFIGURATION.get().render.update(**kwargs)
        yield CONFIGURATION


@contextmanager
def _set_configuration() -> Iterator[ContextVar[GlobalConfiguration]]:
    """Prepares and tidied up the configuration for applying settings.

    This is a context manager, so that a setting can be temporarily overridden and automatically restored.
    """
    try:
        previous = CONFIGURATION.get()
    except LookupError:
        previous = GlobalConfiguration(
            construct=ConstructConfiguration(), render=default_renderer_config()
        )
        CONFIGURATION.set(previous)
    previous = copy.deepcopy(previous)

    try:
        yield CONFIGURATION
    finally:
        CONFIGURATION.set(previous)


@lru_cache
def default_renderer_config() -> RenderConfiguration:
    """Gets the default renderer configuration.

    This may be called frequently, but is cached so note that any changes to the
    wrapped config function will _not_ be reflected during the process.

    It is a light wrapper for `default_config` in the supplier renderer module.

    Returns: the default configuration dict for the chosen renderer.
    """
    try:
        # We have to use a cast as we do not know if the renderer module is valid.
        render_module = cast(
            BaseRenderModule, importlib.import_module("__renderer_mod__")
        )
        default_config: Callable[[], RenderConfiguration] = render_module.default_config
    except ImportError:
        return {}
    return default_config()


@lru_cache
def default_construct_config() -> ConstructConfiguration:
    """Gets the default construct-time configuration.

    This is the primary mechanism for configuring dewret internals, so these defaults
    should be carefully chosen and, if they change, that likely has an impact on backwards compatibility
    from a SemVer perspective.

    Returns: configuration dictionary with default construct values.
    """
    return ConstructConfiguration(
        flatten_all_nested=False,
        allow_positional_args=False,
        allow_plain_dict_fields=False,
        field_separator="/",
        field_index_types="int",
    )


def get_configuration(key: str) -> RawType:
    """Retrieve the configuration or (silently) return the default.

    Helps avoid a proliferation of `set_configuration` calls by not erroring if it has not been called.
    However, the cost is that the user may accidentally put configuration-affected logic outside a
    set_configuration call and be surprised that the behaviour is inexplicibly not as expected.

    Args:
        key: configuration key to retrieve.

    Returns: (preferably) a JSON/YAML-serializable construct.
    """
    try:
        return getattr(CONFIGURATION.get().construct, key)  # type: ignore
    except LookupError:
        # TODO: Not sure what the best way to typehint this is.
        return getattr(ConstructConfiguration(), key)  # type: ignore


def get_render_configuration(key: str) -> RawType:
    """Retrieve configuration for the active renderer.

    Finds the current user-set configuration, defaulting back to the chosen renderer module's declared
    defaults.

    Args:
        key: configuration key to retrieve.

    Returns: (preferably) a JSON/YAML-serializable construct.
    """
    try:
        return CONFIGURATION.get().render.get(key)
    except LookupError:
        return default_renderer_config().get(key)


class WorkflowComponent:
    """Base class for anything directly tied to an individual `Workflow`.

    Attributes:
        __workflow__: the `Workflow` that this is tied to.
    """

    __workflow_real__: WorkflowProtocol

    def __init__(self, *args: Any, workflow: WorkflowProtocol, **kwargs: Any):
        """Tie to a `Workflow`.

        All subclasses must call this.

        Args:
            workflow: the `Workflow` to tie to.
            *args: remainder of arguments for other initializers.
            **kwargs: remainder of arguments for other initializers.
        """
        self.__workflow__ = workflow
        super().__init__(*args, **kwargs)

    @property
    def __workflow__(self) -> WorkflowProtocol:
        """Workflow to which this reference applies."""
        return self.__workflow_real__

    @__workflow__.setter
    def __workflow__(self, workflow: WorkflowProtocol) -> None:
        """Workflow to which this reference applies."""
        self.__workflow_real__ = workflow


class Reference(Generic[U], Symbol, WorkflowComponent):
    """Superclass for all symbolic references to values."""

    _type: type[U] | None = None
    __iterated__: bool = False

    def __init__(self, *args: Any, typ: type[U] | None = None, **kwargs: Any):
        """Extract any specified type.

        Args:
            typ: type to override any inference, or None.
            *args: any other arguments for other initializers (e.g. mixins).
            **kwargs: any other arguments for other initializers (e.g. mixins).
        """
        self._type = typ
        super().__init__(*args, **kwargs)

    @property
    def name(self) -> str:
        """Printable name of the reference."""
        return self.__name__

    def __new__(cls, *args: Any, **kwargs: Any) -> "Reference[U]":
        """As all references are sympy Expressions, the real returned object must be made with Expr."""
        instance = Expr.__new__(cls)
        instance._assumptions0 = {}
        return cast(Reference[U], instance)

    @property
    def __root_name__(self) -> str:
        """Root name on which to suffix/prefix any derived names (with fields, etc.).

        For example, the base name of `add_thing-12345[3]` should be `add_thing`.

        Returns: basic name as a string.
        """
        raise NotImplementedError(
            "Reference must have a '__root_name__' property or override '__name__'"
        )

    @property
    def __type__(self) -> type:
        """Type of the reference target, if known."""
        if self._type is not None:
            return self._type
        raise NotImplementedError()

    def _raise_unevaluatable_error(self) -> None:
        """Convenience method to consistently formulate an UnevaluatableError for this reference."""
        raise UnevaluatableError(
            f"This reference, {self.__name__}, cannot be evaluated during construction."
        )

    def __eq__(self, other: object) -> Any:
        """Test equality at construct-time, if sensible.

        Raises:
            UnevaluatableError: if it seems the user is confusing this with a runtime check.
        """
        if isinstance(other, list) or other is None:
            return False
        if not isinstance(other, Reference) and not isinstance(other, Basic):
            self._raise_unevaluatable_error()
        return super().__eq__(other)

    def __float__(self) -> bool:
        """Catch accidental float casts.

        Raises:
            UnevaluatableError: unconditionally, as it seems the user is confusing this with a runtime check.
        """
        self._raise_unevaluatable_error()
        return False

    def __int__(self) -> bool:
        """Catch accidental int casts.

        Raises:
            UnevaluatableError: unconditionally, as it seems the user is confusing this with a runtime check.
        """
        self._raise_unevaluatable_error()
        return False

    def __bool__(self) -> bool:
        """Catch accidental bool casts.

        Note that this means ambiguous checks such as `if ref: ...` will error, and should be `if ref is None: ...`
        or similar.

        Raises:
            UnevaluatableError: unconditionally, as it seems the user is confusing this with a runtime check.
        """
        self._raise_unevaluatable_error()
        return False

    @property
    def __name__(self) -> str:
        """Referral name for this reference.

        Returns: an internal name to refer to the reference target.
        """
        workflow = self.__workflow__
        name = self.__root_name__
        return workflow.remap(name) if workflow is not None else name

    def __str__(self) -> str:
        """Global description of the reference.

        Returns the _internal_ name.
        """
        return self.__name__


class IterableMixin(Reference[U]):
    """Functionality for iterating over references to give new references."""

    __fixed_len__: int | None = None

    def __init__(self, typ: type[U] | None = None, **kwargs: Any):
        """Extract length, if available from type.

        Captures types of the form (e.g.) `tuple[int, float]` and records the length
        as being (e.g.) 2.
        """
        super().__init__(typ=typ, **kwargs)
        base = strip_annotations(self.__type__)[0]
        if get_origin(base) == tuple and (args := get_args(base)):
            # In the special case of an explicitly-typed tuple, we can state a length.
            self.__fixed_len__ = len(args)

    def __len__(self) -> int:
        """Length of this iterable, if available.

        The two cases that this is likely to be available are if the reference target
        has been type-hinted as a `tuple` with a specific, fixed number of type arguments,
        or if the target has been annotated with `Fixed(...)` indicating that the length
        of the default value can be hard-coded into the output, and therefore that it can
        be used for graph-building logic. The most useful application of this is likely to
        be in for-loops and generators, as we can create variable references for each iteration
        but can nonetheless execute the loop as we know how many iterations occur.

        Returns: length of the iterable, if available.
        """
        if self.__fixed_len__ is None:
            raise TypeError(
                "This iterable reference does not have a known fixed length, "
                "consider using `Fixed[...]` with a default, or typehint using `tuple[int, float]` (or similar) "
                "to tell dewret how long it should be."
            )
        return self.__fixed_len__

    def __iter__(self) -> Generator[Reference[U], None, None]:
        """Execute the iteration.

        Note that this does _not_ return the yielded values of `__inner_iter__`, so that
        it can be used with impunity to do actual iteration, and we will _always_ return
        references here.

        Returns: a generator that will give a new reference for every iteration.
        """
        for count, _ in enumerate(self.__inner_iter__()):
            yield super().__getitem__(count)

    def __inner_iter__(self) -> Generator[Any, None, None]:
        """Overrideable iterator for looping over the wrapped object.

        Returns: a generator that will yield ints if this reference is known to be fixed length, or will go
        forever yielding None otherwise.
        """
        if self.__fixed_len__ is not None:
            for i in range(self.__fixed_len__):
                yield i
        else:
            while True:
                yield None

    def __getitem__(self, attr: str | int) -> "Reference[U] | Any":
        """Get a reference to an individual item/field.

        Args:
            attr: index or fieldname.

        Returns: a reference to the same target as this reference but a level deeper.
        """
        return super().__getitem__(attr)


class IteratedGenerator(Generic[U]):
    """Sentinel value for capturing that an iteration has occured without performing it.

    Allows us to lazily evaluate a loop, for instance, in the renderer. This may be relevant
    if the renderer wishes to postpone iteration to runtime, and simply record it is required,
    rather than evaluating the iterator.
    """

    __wrapped__: Reference[U]

    def __init__(self, to_wrap: Reference[U]):
        """Capture wrapped reference.

        Args:
            to_wrap: reference to wrap.
        """
        self.__wrapped__ = to_wrap

    def __iter__(self) -> Generator[Reference[U], None, None]:
        """Loop through the wrapped reference.

        This will tag the references that are returned, so that the renderer can see this has
        happened.

        Returns: a generator looping over the wrapped reference with a counter as the "field".
        """
        count = -1
        for _ in self.__wrapped__.__inner_iter__():
            ref = self.__wrapped__[(count := count + 1)]
            ref.__iterated__ = True
            yield ref


@dataclass
class Raw:
    """Value object for any raw types.

    This is able to hash raw types consistently and provides
    a single type for validating type-consistency.

    Attributes:
        value: the real value, e.g. a `str`, `int`, ...
    """

    value: RawType

    def __hash__(self) -> int:
        """Provide a hash that is unique to the `value` member."""
        return hash(repr(self))

    def __repr__(self) -> str:
        """Convert to a consistent, string representation."""
        value: str
        if isinstance(self.value, bytes):
            value = base64.b64encode(self.value).decode("ascii")
        else:
            value = str(self.value)
        return f"{type(self.value).__name__}|{value}"



# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Backend base protocol.

Definition of a protocol that valid backend modules must fulfil.
"""

from typing import Protocol, Any
from concurrent.futures import ThreadPoolExecutor
from dewret.workflow import LazyFactory, Lazy, Workflow, StepReference, Target


class BackendModule(Protocol):
    """Requirements for a valid backend module.

    Essential criteria that a backend module must fulfil, such as a
    lazy-evaluation decorator and a task execution function.

    Attributes:
        lazy: Callable that takes a function and returns a lazy-evaluated
            version of it, appropriate to the backend.
    """

    lazy: LazyFactory

    def run(
        self,
        workflow: Workflow | None,
        task: Lazy | list[Lazy] | tuple[Lazy],
        thread_pool: ThreadPoolExecutor | None = None,
    ) -> StepReference[Any] | list[StepReference[Any]] | tuple[StepReference[Any]]:
        """Execute a lazy task for this `Workflow`.

        Args:
            workflow: `Workflow` that is being executed.
            task: task that forms the output.
            thread_pool: the thread pool that should be used for this execution.

        Returns:
            Reference to the final output step.
        """
        ...

    def unwrap(self, lazy: Lazy) -> Target:
        """Unwraps a lazy-evaluated function to get the function.

        Ideally, we could use the `__wrapped__` property but not all
        workflow engines support this, and most importantly, dask has
        only done so as of 2024.03.

        Args:
            lazy: task to be unwrapped.

        Returns:
            Original target.

        Raises:
            RuntimeError: if the task is not a wrapped function.
        """
        ...

    def is_lazy(self, lazy: Any) -> bool:
        """Confirm whether this is a lazy-evaluatable function.

        Args:
            lazy: suspected lazy-evaluatable function to check.

        Returns:
            True if this is a lazy-evaluatable function for this backend, otherwise False.
        """
        ...



"""Test the dask backend."""

import pytest
from typing import cast
from dewret.workflow import Lazy
from dewret.backends.backend_dask import lazy, run


def inc_task(base: int) -> int:
    """Increments a value by one and returns it."""
    return 1 + base


def add_task(left: int, right: int) -> int:
    """Adds two values and returns the result."""
    return left + right


def test_can_run_task() -> None:
    """Check whether a dask task can run via `TaskManager`."""
    task: Lazy = lazy(inc_task)(base=3)
    incremented: int = cast(int, run(None, task))
    assert incremented == 4


def test_missing_arguments_throw_error_in_dask() -> None:
    """Check whether a dask task can run via `TaskManager`."""
    task: Lazy = lazy(add_task)(left=3)
    with pytest.raises(TypeError):
        run(None, task)



# Copyright 2024- Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tooling for managing annotations.

Provides `FunctionAnalyser`, a toolkit that takes a `Callable` and can interrogate it
for annotations, with some intelligent searching beyond the obvious location.
"""

import inspect
import ast
import sys
import importlib
from ._cpython_mod import getclosurevars
from functools import lru_cache
from types import FunctionType, ModuleType
from typing import (
    Any,
    TypeVar,
    Annotated,
    Callable,
    get_origin,
    get_args,
    Mapping,
    get_type_hints,
)

T = TypeVar("T")
AtRender = Annotated[T, "AtRender"]
Fixed = Annotated[T, "Fixed"]


class FunctionAnalyser:
    """Convenience class for analysing a function with reduced duplication of effort.

    Attributes:
        _fn: the wrapped callable
        _annotations: stored annotations for the function.
    """

    _fn: Callable[..., Any]
    _annotations: dict[str, Any]

    def __init__(self, fn: Callable[..., Any]):
        """Set the function.

        If `fn` is a class, it takes the constructor, and if it is a method, it takes
        the `__func__` attribute.
        """
        if inspect.isclass(fn):
            self.fn = fn.__init__
        elif inspect.ismethod(fn):
            self.fn = fn.__func__
        else:
            self.fn = fn

    @property
    def return_type(self) -> Any:
        """Return type of the callable.

        Returns: expected type of the return value.

        Raises:
          ValueError: if the return value does not appear to be type-hinted.
        """
        hints = get_type_hints(inspect.unwrap(self.fn), include_extras=True)
        if "return" not in hints or hints["return"] is None:
            raise ValueError(f"Could not find type-hint for return value of {self.fn}")
        typ = hints["return"]
        return typ

    @staticmethod
    def _typ_has(typ: type, annotation: type) -> bool:
        """Check if the type has an annotation.

        Args:
            typ: type to check.
            annotation: the Annotated to compare against.

        Returns: True if the type has the given annotation, otherwise False.
        """
        if not hasattr(annotation, "__metadata__"):
            return False
        if origin := get_origin(typ):
            if (
                origin is Annotated
                and hasattr(typ, "__metadata__")
                and typ.__metadata__ == annotation.__metadata__
            ):
                return True
        if any(FunctionAnalyser._typ_has(arg, annotation) for arg in get_args(typ)):
            return True
        return False

    def get_all_module_names(self) -> dict[str, Any]:
        """Find all of the annotations within this module."""
        return get_type_hints(sys.modules[self.fn.__module__], include_extras=True)

    def get_all_imported_names(self) -> dict[str, tuple[ModuleType, str]]:
        """Find all of the annotations that were imported into this module."""
        return self._get_all_imported_names(sys.modules[self.fn.__module__])

    @staticmethod
    @lru_cache
    def _get_all_imported_names(mod: ModuleType) -> dict[str, tuple[ModuleType, str]]:
        """Get all of the names with this module, and their original locations.

        Args:
            mod: a module in the `sys.modules`.

        Returns:
            A dict whose keys are the known names in the current module, where the Callable lives,
            and whose values are pairs of the module and the remote name.
        """
        ast_tree = ast.parse(inspect.getsource(mod))
        imported_names = {}
        for node in ast.walk(ast_tree):
            if isinstance(node, ast.ImportFrom):
                for name in node.names:
                    imported_names[name.asname or name.name] = (
                        importlib.import_module(
                            "".join(["."] * node.level) + (node.module or ""),
                            package=mod.__package__,
                        ),
                        name.name,
                    )
        return imported_names

    @property
    def free_vars(self) -> dict[str, Any]:
        """Get the free variables for this Callable."""
        if self.fn.__code__ and self.fn.__closure__:
            return dict(
                zip(
                    self.fn.__code__.co_freevars,
                    (c.cell_contents for c in self.fn.__closure__),
                    strict=False,
                )
            )
        return {}

    def get_argument_annotation(self, arg: str, exhaustive: bool = False) -> Any:
        """Retrieve the annotations for this argument.

        Args:
            arg: name of the argument.
            exhaustive: True if we should search outside the function itself, into the module globals, and into imported modules.

        Returns: annotation if available, else None.
        """
        typ: type | None = None
        if typ := self.fn.__annotations__.get(arg):
            if isinstance(typ, str):
                typ = get_type_hints(self.fn, include_extras=True).get(arg)
        elif exhaustive:
            if anns := get_type_hints(
                sys.modules[self.fn.__module__], include_extras=True
            ):
                if typ := anns.get(arg):
                    ...
                elif orig_pair := self.get_all_imported_names().get(arg):
                    orig_module, orig_name = orig_pair
                    typ = orig_module.__annotations__.get(orig_name)
                elif value := self.free_vars.get(arg):
                    if not inspect.isclass(value) or inspect.isfunction(value):
                        raise RuntimeError(
                            f"Cannot use free variables - please put {arg} at the global scope"
                        )
        return typ

    def argument_has(
        self, arg: str, annotation: type, exhaustive: bool = False
    ) -> bool:
        """Check if the named argument has the given annotation.

        Args:
            arg: argument to retrieve.
            annotation: Annotated to search for.
            exhaustive: whether to check the globals and other modules.

        Returns: True if the Annotated is present in this argument's annotation.
        """
        typ = self.get_argument_annotation(arg, exhaustive)
        return bool(typ and self._typ_has(typ, annotation))

    def is_at_construct_arg(self, arg: str, exhaustive: bool = False) -> bool:
        """Convience function to check for `AtConstruct`, wrapping `FunctionAnalyser.argument_has`."""
        return self.argument_has(arg, AtRender, exhaustive)

    @property
    def globals(self) -> Mapping[str, Any]:
        """Get the globals for this Callable."""
        try:
            fn_tuple = getclosurevars(self.fn)
            fn_globals = dict(fn_tuple.globals)
            fn_globals.update(fn_tuple.nonlocals)
        # This covers the case of wrapping, rather than decorating.
        except TypeError:
            fn_globals = {}
        return fn_globals

    @property
    def unbound(self) -> dict[str, str | None]:
        """Get the globals for this Callable."""
        fn_tuple = getclosurevars(self.fn)
        unbound: dict[str, str | None] = {}
        for var in fn_tuple.unbound:
            if ":" in var:
                ref, impt = var.split(":")
                unbound[ref] = impt
            else:
                unbound[ref] = None
        return unbound

    def with_new_globals(self, new_globals: dict[str, Any]) -> Callable[..., Any]:
        """Create a Callable that will run the current Callable with new globals."""
        code = self.fn.__code__
        fn_name = self.fn.__name__
        all_globals = dict(self.globals)
        all_globals.update(new_globals)
        return FunctionType(
            code,
            all_globals,
            name=fn_name,
            closure=self.fn.__closure__,
            argdefs=self.fn.__defaults__,
        )



"""Verify CWL can be made with parameters."""

import yaml
from dewret.tasks import task, construct
from dewret.workflow import param
from dewret.renderers.cwl import render

from ._lib.extra import double, sum

INPUT_NUM = 3


@task()
def rotate(num: int) -> int:
    """Rotate an integer."""
    return (num + INPUT_NUM) % INPUT_NUM


def test_cwl_parameters() -> None:
    """Check whether we can spot input parameters.

    Produces CWL that reference input parameters based on local/global variables.
    """
    result = rotate(num=3)
    workflow = construct(result, simplify_ids=True)
    rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          INPUT_NUM:
            label: INPUT_NUM
            type: int
            default: 3
          rotate-1-num:
            label: num
            type: int
            default: 3
        outputs:
          out:
            label: out
            outputSource: rotate-1/out
            type: int
        steps:
          rotate-1:
            run: rotate
            in:
                num:
                    source: rotate-1-num
                INPUT_NUM:
                    source: INPUT_NUM
            out: [out]
    """)


def test_complex_parameters() -> None:
    """Check whether we can link between multiple steps and have parameters.

    Produces CWL that has references between multiple steps.
    """
    num = param("numx", 23)
    result = sum(left=double(num=rotate(num=num)), right=rotate(num=rotate(num=23)))
    workflow = construct(result, simplify_ids=True)
    rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          INPUT_NUM:
            label: INPUT_NUM
            type: int
            default: 3
          numx:
            label: numx
            type: int
            default: 23
          rotate-2-num:
            label: num
            type: int
            default: 23
        outputs:
          out:
            label: out
            outputSource: sum-1/out
            type: [int, float]
        steps:
          rotate-1:
            run: rotate
            in:
                INPUT_NUM:
                    source: INPUT_NUM
                num:
                    source: rotate-2/out
            out: [out]
          double-1:
            run: double
            in:
                num:
                    source: rotate-3/out
            out: [out]
          rotate-2:
            run: rotate
            in:
                INPUT_NUM:
                    source: INPUT_NUM
                num:
                    source: rotate-2-num
            out: [out]
          rotate-3:
            run: rotate
            in:
                INPUT_NUM:
                    source: INPUT_NUM
                num:
                    source: numx
            out: [out]
          sum-1:
            run: sum
            in:
                left:
                    source: double-1/out
                right:
                    source: rotate-1/out
            out: [out]
    """)



"""Verify CWL output is OK."""

import yaml
import pytest
from datetime import datetime, timedelta
from dewret.core import set_configuration
from dewret.tasks import construct, task, factory, TaskException
from dewret.renderers.cwl import render
from dewret.utils import hasher
from dewret.workflow import param

from ._lib.extra import (
    pi,
    increment,
    double,
    mod10,
    sum,
    triple_and_one,
    tuple_float_return,
)


@task()
def floor(num: int | float) -> int:
    """Converts int/float to int."""
    return int(num)


@task()
def days_in_future(now: datetime, num: int | float) -> datetime:
    """Add `num` days to `now`.

    Args:
        now: current datetime.
        num: count of days.
    """
    return now + timedelta(days=num)


def test_basic_cwl() -> None:
    """Check whether we can produce simple CWL.

    Produces simplest possible CWL from a workflow, using
    a pure function.
    """
    result = pi()
    workflow = construct(result)
    rendered = render(workflow)["__root__"]
    hsh = hasher(("pi",))

    assert rendered == yaml.safe_load(f"""
        cwlVersion: 1.2
        class: Workflow
        inputs: {{}}
        outputs:
          out:
            label: out
            outputSource: pi-{hsh}/out
            type: float
        steps:
          pi-{hsh}:
            run: pi
            in: {{}}
            out: [out]
    """)


def test_input_factories() -> None:
    """Use input factories to input complex types.

    Tests whether input factories can be treated as steps,
    or complex input, as a flag choice to the renderer.
    """

    def get_now() -> datetime:
        return datetime.now()

    now = factory(get_now)()
    result = days_in_future(now=now, num=3)
    workflow = construct(result, simplify_ids=True)
    rendered = render(workflow, allow_complex_types=True, factories_as_params=True)[
        "__root__"
    ]

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          days_in_future-1-num:
            default: 3
            type: int
            label: num
          get_now-1:
            label: get_now-1
            type: datetime
        outputs:
          out:
            label: out
            outputSource: days_in_future-1/out
            type: datetime
        steps:
          days_in_future-1:
            run: days_in_future
            in:
              num:
                source: days_in_future-1-num
              now:
                source: get_now-1/out
            out: [out]
    """)

    rendered = render(workflow, allow_complex_types=True)["__root__"]

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          days_in_future-1-num:
            default: 3
            type: int
            label: num
        outputs:
          out:
            label: out
            outputSource: days_in_future-1/out
            type: datetime
        steps:
          days_in_future-1:
            run: days_in_future
            in:
              num:
                source: days_in_future-1-num
              now:
                source: get_now-1/out
            out: [out]
          get_now-1:
            in: {}
            out: [out]
            run: get_now
    """)


def test_cwl_with_parameter() -> None:
    """Check whether we can move raw input to parameters.

    Produces CWL for a call with a changeable raw value, that is converted
    to a parameter, if and only if we are calling from outside a subworkflow.
    """
    result = increment(num=3)
    workflow = construct(result)
    rendered = render(workflow)["__root__"]
    num_param = list(workflow.find_parameters())[0]
    hsh = hasher(("increment", ("num", f"int|:param:{num_param._.unique_name}")))

    assert rendered == yaml.safe_load(f"""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          increment-{hsh}-num:
            label: num
            type: int
            default: 3
        outputs:
          out:
            label: out
            outputSource: increment-{hsh}/out
            type: int
        steps:
          increment-{hsh}:
            run: increment
            in:
                num:
                    source: increment-{hsh}-num
            out: [out]
    """)


def test_cwl_with_positional_parameter() -> None:
    """Check whether we can move raw input to parameters.

    Produces CWL for a call with a changeable raw value, that is converted
    to a parameter, if and only if we are calling from outside a subworkflow.
    """
    with pytest.raises(TaskException) as _:
        result = increment(3)
    with set_configuration(allow_positional_args=True):
        result = increment(3)
        workflow = construct(result)
        rendered = render(workflow)["__root__"]
    num_param = list(workflow.find_parameters())[0]
    hsh = hasher(("increment", ("num", f"int|:param:{num_param._.unique_name}")))

    assert rendered == yaml.safe_load(f"""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          increment-{hsh}-num:
            label: num
            type: int
            default: 3
        outputs:
          out:
            label: out
            outputSource: increment-{hsh}/out
            type: int
        steps:
          increment-{hsh}:
            run: increment
            in:
                num:
                    source: increment-{hsh}-num
            out: [out]
    """)


def test_cwl_without_default() -> None:
    """Check whether we can produce CWL without a default value.

    Uses a manually created parameter to avoid a default.
    """
    my_param = param("my_param", typ=int)

    result = increment(num=my_param)
    workflow = construct(result)
    rendered = render(workflow)["__root__"]
    hsh = hasher(("increment", ("num", "int|:param:my_param")))

    assert rendered == yaml.safe_load(f"""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          my_param:
            label: my_param
            type: int
        outputs:
          out:
            label: out
            outputSource: increment-{hsh}/out
            type: int
        steps:
          increment-{hsh}:
            run: increment
            in:
                num:
                    source: my_param
            out: [out]
    """)


def test_cwl_with_subworkflow() -> None:
    """Check whether we can produce a subworkflow from CWL."""
    my_param = param("num", typ=int)
    result = increment(num=floor(num=triple_and_one(num=increment(num=my_param))))
    workflow = construct(result, simplify_ids=True)
    subworkflows = render(workflow)
    rendered = subworkflows["__root__"]
    del subworkflows["__root__"]

    assert len(subworkflows) == 1
    assert isinstance(subworkflows, dict)
    name, subworkflow = list(subworkflows.items())[0]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: increment-1/out
            type: int
        steps:
          floor-1:
             in:
               num:
                 source: triple_and_one-1/out
             out: [out]
             run: floor
          increment-2:
             in:
               num:
                 source: num
             out: [out]
             run: increment
          increment-1:
             in:
               num:
                 source: floor-1/out
             out: [out]
             run: increment
          triple_and_one-1:
             in:
               num:
                 source: increment-2/out
             out: [out]
             run: triple_and_one
    """)

    assert subworkflow == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: sum-1-1/out
            type:
            - int
            - float
        steps:
          double-1-1:
            in:
              num:
                source: num
            out:
            - out
            run: double
          sum-1-2:
            in:
              left:
                source: double-1-1/out
              right:
                source: num
            out:
            - out
            run: sum
          sum-1-1:
            in:
              left:
                source: sum-1-2/out
              right:
                default: 1
            out:
            - out
            run: sum
    """)


def test_cwl_references() -> None:
    """Check whether we can link between steps.

    Produces CWL that has references between steps.
    """
    result = double(num=increment(num=3))
    workflow = construct(result)
    rendered = render(workflow)["__root__"]
    num_param = list(workflow.find_parameters())[0]
    hsh_increment = hasher(
        ("increment", ("num", f"int|:param:{num_param._.unique_name}"))
    )
    hsh_double = hasher(("double", ("num", f"increment-{hsh_increment}")))

    assert rendered == yaml.safe_load(f"""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          increment-{hsh_increment}-num:
            label: num
            type: int
            default: 3
        outputs:
          out:
            label: out
            outputSource: double-{hsh_double}/out
            type:
            - int
            - float
        steps:
          increment-{hsh_increment}:
            run: increment
            in:
                num:
                    source: increment-{hsh_increment}-num
            out: [out]
          double-{hsh_double}:
            run: double
            in:
                num:
                    source: increment-{hsh_increment}/out
            out: [out]
    """)


def test_complex_cwl_references() -> None:
    """Check whether we can link between multiple steps.

    Produces CWL that has references between multiple steps.
    """
    result = sum(left=double(num=increment(num=23)), right=mod10(num=increment(num=23)))
    workflow = construct(result, simplify_ids=True)
    rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          increment-1-num:
            label: num
            type: int
            default: 23
        outputs:
          out:
            label: out
            outputSource: sum-1/out
            type:
            - int
            - float
        steps:
          increment-1:
            run: increment
            in:
                num:
                    source: increment-1-num
            out: [out]
          double-1:
            run: double
            in:
                num:
                    source: increment-1/out
            out: [out]
          mod10-1:
            run: mod10
            in:
                num:
                    source: increment-1/out
            out: [out]
          sum-1:
            run: sum
            in:
                left:
                    source: double-1/out
                right:
                    source: mod10-1/out
            out: [out]
    """)


def test_cwl_with_subworkflow_and_raw_params() -> None:
    """Check whether we can produce a subworkflow from CWL."""
    my_param = param("num", typ=int)
    result = increment(num=floor(num=triple_and_one(num=sum(left=my_param, right=3))))
    workflow = construct(result, simplify_ids=True)
    subworkflows = render(workflow)
    rendered = subworkflows["__root__"]

    del subworkflows["__root__"]
    assert len(subworkflows) == 1
    assert isinstance(subworkflows, dict)
    name, subworkflow = list(subworkflows.items())[0]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
          sum-1-right:
            default: 3
            label: right
            type: int
        outputs:
          out:
            label: out
            outputSource: increment-1/out
            type: int
        steps:
          floor-1:
             in:
               num:
                 source: triple_and_one-1/out
             out: [out]
             run: floor
          increment-1:
             in:
               num:
                 source: floor-1/out
             out: [out]
             run: increment
          sum-1:
             in:
               left:
                 source: num
               right:
                 source: sum-1-right
             out: [out]
             run: sum
          triple_and_one-1:
             in:
               num:
                 source: sum-1/out
             out: [out]
             run: triple_and_one
    """)

    assert subworkflow == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: [
              int,
              float
            ]
        outputs:
          out:
            label: out
            outputSource: sum-1-1/out
            type:
            - int
            - float
        steps:
          double-1-1:
            in:
              num:
                source: num
            out:
            - out
            run: double
          sum-1-2:
            in:
              left:
                source: double-1-1/out
              right:
                source: num
            out:
            - out
            run: sum
          sum-1-1:
            in:
              left:
                source: sum-1-2/out
              right:
                default: 1
            out:
            - out
            run: sum
    """)


def test_tuple_floats() -> None:
    """Checks whether we can return a tuple.

    Produces CWL that has a tuple of 2 values of type float.
    """
    result = tuple_float_return()
    workflow = construct(result, simplify_ids=True)
    rendered = render(workflow)["__root__"]
    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs: {}
        outputs:
          out:
            label: out
            outputSource: tuple_float_return-1/out
            items:
              - float
              - float
            type: array
        steps:
          tuple_float_return-1:
            run: tuple_float_return
            in: {}
            out: [out]
    """)



"""Check configuration is consistent and usable."""

import yaml
import pytest
from dewret.tasks import construct, workflow, TaskException
from dewret.renderers.cwl import render
from dewret.core import set_configuration
from dewret.annotations import AtRender
from ._lib.extra import increment


@workflow()
def floor(num: int, expected: AtRender[bool]) -> int:
    """Converts int/float to int."""
    from dewret.core import get_configuration

    if get_configuration("flatten_all_nested") != expected:
        raise AssertionError(
            f"Not expected configuration: {str(get_configuration('flatten_all_nested'))} != {expected}"
        )
    return increment(num=num)


def test_cwl_with_parameter() -> None:
    """Test workflows with configuration parameters."""
    with set_configuration(flatten_all_nested=True):
        result = increment(num=floor(num=3, expected=True))
        workflow = construct(result, simplify_ids=True)

    with (
        pytest.raises(TaskException) as exc,
        set_configuration(flatten_all_nested=False),
    ):
        result = increment(num=floor(num=3, expected=True))
        workflow = construct(result, simplify_ids=True)
    assert "AssertionError" in str(exc.getrepr())

    with set_configuration(flatten_all_nested=True):
        result = increment(num=floor(num=3, expected=True))
        workflow = construct(result, simplify_ids=True)
    rendered = render(workflow)["__root__"]
    num_param = list(workflow.find_parameters())[0]
    assert num_param

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          num:
            label: num
            type: int
            default: 3
        outputs:
          out:
            label: out
            outputSource: increment-1/out
            type: int
        steps:
          increment-2:
            run: increment
            in:
                num:
                    source: num
            out: [out]
          increment-1:
            run: increment
            in:
                num:
                    source: increment-2/out
            out: [out]
    """)



"""Check subworkflow behaviour is as expected."""

from typing import Callable
from queue import Queue
import yaml
from dewret.tasks import construct, workflow, task, factory
from dewret.core import set_configuration
from dewret.renderers.cwl import render
from dewret.workflow import param
from attrs import define

from ._lib.extra import increment, sum, pi

CONSTANT: int = 3

QueueFactory: Callable[..., Queue[int]] = factory(Queue)

GLOBAL_QUEUE: Queue[int] = QueueFactory()


@task()
def pop(queue: Queue[int]) -> int:
    """Remove element of a queue."""
    return queue.get()


@task()
def to_int(num: int | float) -> int:
    """Cast to an int."""
    return int(num)


@task()
def add_and_queue(num: int, queue: Queue[int]) -> Queue[int]:
    """Add a global constant to a number."""
    queue.put(num)
    return queue


@workflow()
def make_queue(num: int | float) -> Queue[int]:
    """Add a number to a queue."""
    queue = QueueFactory()
    return add_and_queue(num=to_int(num=num), queue=queue)


@workflow()
def get_global_queue(num: int | float) -> Queue[int]:
    """Add a number to a global queue."""
    return add_and_queue(num=to_int(num=num), queue=GLOBAL_QUEUE)


@workflow()
def get_global_queues(num: int | float) -> list[Queue[int] | int]:
    """Add a number to a global queue."""
    return [
        add_and_queue(num=to_int(num=num), queue=GLOBAL_QUEUE),
        add_constant(num=num),
    ]


@workflow()
def add_constant(num: int | float) -> int:
    """Add a global constant to a number."""
    return to_int(num=sum(left=num, right=CONSTANT))


@workflow()
def add_constants(num: int | float) -> int:
    """Add a global constant to a number."""
    return to_int(num=sum(left=sum(left=num, right=CONSTANT), right=CONSTANT))


@workflow()
def get_values(num: int | float) -> tuple[int | float, int]:
    """Add a global constant to a number."""
    return (sum(left=num, right=CONSTANT), add_constant(CONSTANT))


def test_cwl_for_pairs() -> None:
    """Check whether we can produce CWL of pairs."""

    @workflow()
    def pair_pi() -> tuple[float, float]:
        return pi(), pi()

    with set_configuration(flatten_all_nested=True):
        result = pair_pi()
        wkflw = construct(result, simplify_ids=True)
    rendered = render(wkflw)["__root__"]

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs: {}
        outputs: [
          {
            label: out,
            outputSource: pi-1/out,
            type: float
          },
          {
            label: out,
            outputSource: pi-1/out,
            type: float
          }
        ]
        steps:
          pi-1:
            run: pi
            in: {}
            out: [out]
    """)


def test_subworkflows_can_use_globals() -> None:
    """Produce a subworkflow that uses a global."""
    my_param = param("num", typ=int)
    result = increment(num=add_constant(num=increment(num=my_param)))
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw)
    rendered = subworkflows["__root__"]

    assert len(subworkflows) == 2
    assert isinstance(subworkflows, dict)

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          CONSTANT:
            label: CONSTANT
            default: 3
            type: int
          num:
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: increment-1/out
            type: int
        steps:
          increment-2:
             in:
               num:
                 source: num
             out: [out]
             run: increment
          increment-1:
             in:
               num:
                 source: add_constant-1/out
             out: [out]
             run: increment
          add_constant-1:
             in:
               CONSTANT:
                 source: CONSTANT
               num:
                 source: increment-2/out
             out: [out]
             run: add_constant
    """)


def test_subworkflows_can_use_factories() -> None:
    """Produce a subworkflow that uses a factory."""
    my_param = param("num", typ=int)
    result = pop(queue=make_queue(num=increment(num=my_param)))
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    rendered = subworkflows["__root__"]

    assert len(subworkflows) == 2
    assert isinstance(subworkflows, dict)

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: pop-1/out
            type: int
        steps:
          increment-1:
             in:
               num:
                 source: num
             out: [out]
             run: increment
          make_queue-1:
             in:
               num:
                 source: increment-1/out
             out: [out]
             run: make_queue
          pop-1:
             in:
               queue:
                 source: make_queue-1/out
             out: [out]
             run: pop
    """)


def test_subworkflows_can_use_global_factories() -> None:
    """Check whether we can produce a subworkflow that uses a global factory."""
    my_param = param("num", typ=int)
    result = pop(queue=get_global_queue(num=increment(num=my_param)))
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    rendered = subworkflows["__root__"]

    assert len(subworkflows) == 2
    assert isinstance(subworkflows, dict)

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
          GLOBAL_QUEUE:
            label: GLOBAL_QUEUE
            type: Queue
        outputs:
          out:
            label: out
            outputSource: pop-1/out
            type: int
        steps:
          increment-1:
             in:
               num:
                 source: num
             out: [out]
             run: increment
          get_global_queue-1:
             in:
               num:
                 source: increment-1/out
               GLOBAL_QUEUE:
                 source: GLOBAL_QUEUE
             out: [out]
             run: get_global_queue
          pop-1:
             in:
               queue:
                 source: get_global_queue-1/out
             out: [out]
             run: pop
    """)


def test_subworkflows_can_return_lists() -> None:
    """Check whether we can produce a subworkflow that returns a list."""
    my_param = param("num", typ=int)
    result = get_global_queues(num=increment(num=my_param))
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    rendered = subworkflows["__root__"]
    del subworkflows["__root__"]

    assert len(subworkflows) == 2
    assert isinstance(subworkflows, dict)
    osubworkflows = sorted(list(subworkflows.items()))

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
          CONSTANT:
            label: CONSTANT
            default: 3
            type: int
          GLOBAL_QUEUE:
            label: GLOBAL_QUEUE
            type: Queue
        outputs:
          out:
            label: out
            items:
            - Queue
            - int
            outputSource: get_global_queues-1/out
            type: array
        steps:
          increment-1:
             in:
               num:
                 source: num
             out: [out]
             run: increment
          get_global_queues-1:
             in:
               num:
                 source: increment-1/out
               CONSTANT:
                 source: CONSTANT
               GLOBAL_QUEUE:
                 source: GLOBAL_QUEUE
             out: [out]
             run: get_global_queues
    """)

    assert osubworkflows[0] == (
        "add_constant-1-1",
        yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
          CONSTANT:
            default: 3
            label: CONSTANT
            type: int
        outputs:
          out:
            label: out
            outputSource: to_int-1-1-1/out
            type: int
        steps:
          sum-1-1-1:
            in:
              left:
                source: num
              right:
                source: CONSTANT
            out:
            - out
            run: sum
          to_int-1-1-1:
            in:
              num:
                source: sum-1-1-1/out
            out:
            - out
            run: to_int
    """),
    )

    assert osubworkflows[1] == (
        "get_global_queues-1",
        yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          CONSTANT:
            default: 3
            label: CONSTANT
            type: int
          num:
            label: num
            type: int
          GLOBAL_QUEUE:
            label: GLOBAL_QUEUE
            type: Queue
        outputs:
          - label: out
            outputSource: add_and_queue-1-1/out
            type: Queue
          - label: out
            outputSource: add_constant-1-1/out
            type: int
        steps:
          add_and_queue-1-1:
            in:
              num:
                source: to_int-1-1/out
              queue:
                source: GLOBAL_QUEUE
            out:
            - out
            run: add_and_queue
          add_constant-1-1:
            in:
              CONSTANT:
                source: CONSTANT
              num:
                source: num
            out:
            - out
            run: add_constant
          to_int-1-1:
            in:
              num:
                source: num
            out:
            - out
            run: to_int
    """),
    )


def test_can_merge_workflows() -> None:
    """Check whether we can merge workflows."""
    my_param = param("num", typ=int)
    value = to_int(num=increment(num=my_param))
    result = sum(left=value, right=increment(num=value))
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    rendered = subworkflows["__root__"]
    del subworkflows["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: sum-1/out
            type: [
              int,
              float
            ]
        steps:
          increment-1:
             in:
               num:
                 source: num
             out: [out]
             run: increment
          increment-2:
             in:
               num:
                 source: to_int-1/out
             out: [out]
             run: increment
          sum-1:
             in:
               left:
                 source: to_int-1/out
               right:
                 source: increment-2/out
             out: [out]
             run: sum
          to_int-1:
             in:
               num:
                 source: increment-1/out
             out: [out]
             run: to_int
    """)


def test_subworkflows_can_use_globals_in_right_scope() -> None:
    """Produce a subworkflow that uses a global."""
    my_param = param("num", typ=int)
    result = increment(num=add_constants(num=increment(num=my_param)))
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw)
    rendered = subworkflows["__root__"]
    del subworkflows["__root__"]

    assert len(subworkflows) == 1
    assert isinstance(subworkflows, dict)
    osubworkflows = sorted(list(subworkflows.items()))

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          CONSTANT:
            default: 3
            label: CONSTANT
            type: int
          num:
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: increment-1/out
            type: int
        steps:
          increment-1:
             in:
               num:
                 source: add_constants-1/out
             out: [out]
             run: increment
          increment-2:
             in:
               num:
                 source: num
             out: [out]
             run: increment
          add_constants-1:
             in:
               CONSTANT:
                 source: CONSTANT
               num:
                 source: increment-2/out
             out: [out]
             run: add_constants
    """)

    assert osubworkflows[0] == (
        "add_constants-1",
        yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          num:
            label: num
            type: int
          CONSTANT:
            default: 3
            label: CONSTANT
            type: int
        outputs:
          out:
            label: out
            outputSource: to_int-1-1/out
            type: int
        steps:
          sum-1-1:
            in:
              left:
                source: num
              right:
                source: CONSTANT
            out:
            - out
            run: sum
          sum-1-2:
            in:
              left:
                source: sum-1-1/out
              right:
                source: CONSTANT
            out:
            - out
            run: sum
          to_int-1-1:
            in:
              num:
                source: sum-1-2/out
            out:
            - out
            run: to_int
    """),
    )


@define
class PackResult:
    """A class representing the counts of card suits in a deck, including hearts, clubs, spades, and diamonds."""

    hearts: int
    clubs: int
    spades: int
    diamonds: int


def test_combining_attrs_and_factories() -> None:
    """Check combining attributes from a dataclass with factory-produced instances."""
    Pack = factory(PackResult)

    @task()
    def sum(left: int, right: int) -> int:
        return left + right

    @workflow()
    def black_total(pack: PackResult) -> int:
        return sum(left=pack.spades, right=pack.clubs)

    pack = Pack(hearts=13, spades=13, diamonds=13, clubs=13)
    wkflw = construct(black_total(pack=pack), simplify_ids=True)
    cwl = render(wkflw, allow_complex_types=True, factories_as_params=True)
    assert cwl["__root__"] == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          PackResult-1:
            label: PackResult-1
            type: record
        outputs:
          out:
            label: out
            outputSource: black_total-1/out
            type: int
        steps:
          black_total-1:
            in:
              pack:
                source: PackResult-1/out
            out:
            - out
            run: black_total
    """)

    assert cwl["black_total-1"] == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          pack:
            label: pack
            type: record
        outputs:
          out:
            label: out
            outputSource: sum-1-1/out
            type: int
        steps:
          sum-1-1:
            in:
              left:
                source: pack/spades
              right:
                source: pack/clubs
            out:
            - out
            run: sum
    """)



"""Testing example renderer.

Correctly fails to import to show a broken module being handled.
"""

from typing import Unpack, TypedDict
from dewret.workflow import Workflow

from .extra import JUMP


class NonfrenderRendererConfiguration(TypedDict):
    allow_complex_types: bool


# This should fail to load as default_config is not present. However it would
# ignore the fact that the return type is not a (subtype of) dict[str, RawType]
# def default_config() -> int:
#     return 3


def render_raw(
    workflow: Workflow, **kwargs: Unpack[NonfrenderRendererConfiguration]
) -> dict[str, str]:
    return {"JUMP": str(JUMP)}



"""Verify CWL can be made with split up and nested calls."""

import yaml
from dewret.tasks import workflow, construct
from dewret.core import set_configuration
from dewret.renderers.cwl import render
from ._lib.extra import double, sum, increase

STARTING_NUMBER: int = 23


@workflow()
def algorithm() -> int | float:
    """Creates a graph of task calls."""
    left = double(num=increase(num=STARTING_NUMBER))
    num = increase(num=17)
    right = increase(num=num)
    return sum(left=left, right=right)


def test_subworkflow() -> None:
    """Check whether we can link between multiple steps and have parameters.

    Produces CWL that has references between multiple steps.
    """
    with set_configuration(flatten_all_nested=True):
        workflow = construct(algorithm(), simplify_ids=True)
        rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          JUMP:
            label: JUMP
            type: float
            default: 1.0
          STARTING_NUMBER:
            default: 23
            label: STARTING_NUMBER
            type: int
        outputs:
          out:
            label: out
            outputSource: sum-1/out
            type: [int, float]
        steps:
          increase-1:
            run: increase
            in:
                JUMP:
                    source: JUMP
                num:
                    source: STARTING_NUMBER
            out: [out]
          increase-2:
            run: increase
            in:
                JUMP:
                    source: JUMP
                num:
                    source: increase-3/out
            out: [out]
          increase-3:
            run: increase
            in:
                JUMP:
                    source: JUMP
                num:
                    default: 17
            out: [out]
          double-1:
            run: double
            in:
                num:
                    source: increase-1/out
            out: [out]
          sum-1:
            run: sum
            in:
                left:
                    source: double-1/out
                right:
                    source: increase-2/out
            out: [out]
    """)



# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module.

General types and functions to centralize common logic.
"""

import hashlib
import json
import sys
import importlib
import importlib.util
from types import FrameType, TracebackType, UnionType, ModuleType
from typing import Any, cast, Protocol, ClassVar, Callable, Iterable, get_args, Hashable
from pathlib import Path
from collections.abc import Sequence, Mapping
from dataclasses import asdict, is_dataclass
from sympy import Basic, Integer, Float, Rational

from .core import Reference, RawType, FirmType, Raw


class Unset:
    """Unset variable, with no default value."""


class DataclassProtocol(Protocol):
    """Format of a dataclass.

    Since dataclasses do not expose a proper type, we use this to
    represent them.
    """

    __dataclass_fields__: ClassVar[dict[str, Any]]


def make_traceback(skip: int = 2) -> TracebackType | None:
    """Creates a traceback for the current frame.

    Necessary to allow tracebacks to be prepped for
    potential errors in lazy-evaluated functions.

    Args:
        skip: number of frames to skip before starting traceback.
    """
    frame: FrameType | None = sys._getframe(skip)
    tb = None
    while frame:
        tb = TracebackType(tb, frame, frame.f_lasti, frame.f_lineno)
        frame = frame.f_back
    return tb


def load_module_or_package(target_name: str, path: Path) -> ModuleType:
    """Convenience loader for modules.

    If an `__init__.py` is found in the same location as the target, it will try to load the renderer module
    as if it is contained in a package and, if it cannot, will fall back to loading the single file.

    Args:
        target_name: module name that should appear in `sys.modules`.
        path: location of the module.

    Returns: the loaded module.
    """
    module: None | ModuleType = None
    exception: None | Exception = None
    package_init = path.parent / "__init__.py"
    # Try to import the module as a package, if possible, to allow relative imports.
    if package_init.exists():
        try:
            spec = importlib.util.spec_from_file_location(
                target_name, str(package_init)
            )
            if spec is None or spec.loader is None:
                raise ImportError(f"Could not open {path.parent} package")
            package = importlib.util.module_from_spec(spec)
            sys.modules[target_name] = package
            spec.loader.exec_module(package)
            module = importlib.import_module(f"{target_name}.{path.stem}", target_name)
        except ImportError as exc:
            exception = exc

    if module is None:
        try:
            spec = importlib.util.spec_from_file_location(target_name, str(path))
            if spec is None or spec.loader is None:
                raise ImportError(f"Could not open {path} module")
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        except ImportError as exc:
            if exception:
                raise exc from exception
            raise exc

    return module


def flatten_if_set(value: Any) -> RawType | Unset:
    """Takes a Raw-like structure and makes it RawType or Unset.

    Flattens if the value is set, but otherwise returns the unset
    sentinel value as-is.

    Args:
        value: value to squash
    """
    if isinstance(value, Unset):
        return value
    return crawl_raw(value)


def crawl_raw(value: Any, action: Callable[[Any], Any] | None = None) -> RawType:
    """Takes a Raw-like structure and makes it RawType.

    Particularly useful for squashing any TypedDicts.

    Args:
        value: value to squash
        action: an callback to apply to each found entry, or None.

    Returns: a structuure that is guaranteed to be raw.

    Raises: RuntimeError if it cannot convert the value to raw.
    """
    if action is not None:
        value = action(value)

    if value is None:
        return value
    if isinstance(value, str) or isinstance(value, bytes):
        return value
    if isinstance(value, Mapping):
        return {key: crawl_raw(item, action) for key, item in value.items()}
    if is_dataclass(value) and not isinstance(value, type):
        return crawl_raw(asdict(value), action)
    if isinstance(value, Sequence):
        return [crawl_raw(item, action) for item in value]
    if (raw := ensure_raw(value)) is not None:
        return raw
    raise RuntimeError(f"Could not flatten: {value}")


def firm_to_raw(value: FirmType) -> RawType:
    """Convenience wrapper for firm structures.

    Turns structures that would be raw, except for tuples, into raw structures
    by mapping any tuples to lists.

    Args:
        value: a firm structure (contains raw/tuple values).

    Returns: a raw structure.
    """
    return crawl_raw(
        value, lambda entry: list(entry) if isinstance(entry, tuple) else entry
    )


def is_expr(value: Any, permitted_references: type = Reference) -> bool:
    """Confirms whether a structure has only raw or expression types.

    Args:
        value: a crawlable structure.
        permitted_references: a class representing the allowed types of References.

    Returns: True if valid, otherwise False.
    """
    return is_raw(
        value,
        lambda x: isinstance(x, Basic)
        or isinstance(x, tuple)
        or isinstance(x, permitted_references)
        or isinstance(x, Raw),
    )


def is_raw_type(typ: type) -> bool:
    """Check if a type counts as "raw"."""
    if isinstance(typ, UnionType):
        return all(is_raw_type(t) for t in get_args(typ))
    return issubclass(typ, str | float | bool | bytes | int | None | list | dict)


def is_firm(value: Any, check: Callable[[Any], bool] | None = None) -> bool:
    """Confirms whether a function is firm.

    That is, whether its contents are raw or tuples.

    Args:
        value: value to check.
        check: any additional check to apply.

    Returns: True if is firm, else False.
    """
    return is_raw(value, lambda x: isinstance(x, tuple) and (check is None or check(x)))


def is_raw(value: Any, check: Callable[[Any], bool] | None = None) -> bool:
    """Check if a variable counts as "raw".

    This works around a checking issue that isinstance of a union of types
    assigned to a variable, such as `RawType`, may throw errors even though Python 3.11+
    does not. Instead, we explicitly make the full union in the statement below.
    """
    # Ideally this would be:
    # isinstance(value, RawType | list[RawType] | dict[str, RawType])
    # but recursive types are problematic.
    if isinstance(
        value, str | float | bool | bytes | int | None | Integer | Float | Rational
    ):
        return True

    if check is not None and check(value):
        return True

    if isinstance(value, Mapping):
        return (
            (isinstance(value, dict) or (check is not None and check(value)))
            and all(is_raw(key, check) for key in value.keys())
            and all(is_raw(val, check) for val in value.values())
        )

    if isinstance(value, Iterable):
        return (
            isinstance(value, list) or (check is not None and check(value))
        ) and all(is_raw(key, check) for key in value)

    return False


def ensure_raw(value: Any, cast_tuple: bool = False) -> RawType | None:
    """Check if a variable counts as "raw".

    This works around a checking issue that isinstance of a union of types
    assigned to a variable, such as `RawType`, may throw errors even though Python 3.11+
    does not. Instead, we explicitly make the full union in the statement below.
    """
    # See is_raw:
    # isinstance(var, RawType | list[RawType] | dict[str, RawType])
    return cast(RawType, value) if is_raw(value) else None


def hasher(construct: FirmType) -> str:
    """Consistently hash a RawType or tuple structure.

    Turns a possibly nested structure of basic types, dicts, lists and tuples
    into a consistent hash.

    Args:
        construct: structure to hash.

    Returns:
        Hash string that should be unique to the construct. The limits of this uniqueness
        have not yet been explicitly calculated.
    """

    def _make_hashable(construct: FirmType) -> Hashable:
        hashed_construct: tuple[Hashable, ...]
        if isinstance(construct, Sequence) and not isinstance(construct, bytes | str):
            if isinstance(construct, Mapping):
                hashed_construct = tuple(
                    (k, hasher(v)) for k, v in sorted(construct.items())
                )
            else:
                # Cast to workaround recursive type
                hashed_construct = tuple(_make_hashable(v) for v in construct)
            return hashed_construct
        if not isinstance(construct, Hashable):
            raise TypeError("Could not hash arguments")
        return construct

    if isinstance(construct, Hashable):
        hashed_construct = construct
    else:
        hashed_construct = _make_hashable(construct)
    construct_as_string = json.dumps(hashed_construct)
    hsh = hashlib.md5()
    hsh.update(construct_as_string.encode())
    return hsh.hexdigest()



"""Verify we can interrogate annotations."""

import pytest
import yaml

from dewret.tasks import construct, workflow, TaskException
from dewret.renderers.cwl import render
from dewret.annotations import AtRender, FunctionAnalyser, Fixed
from dewret.core import set_configuration

from ._lib.extra import increment, sum, try_nothing

from typing import Any

ARG1: AtRender[bool] = True
ARG2: bool = False


class MyClass:
    """A mock class to wrap values and mock processing of data."""

    def method(self, arg1: bool, arg2: AtRender[int]) -> float:
        """A mock method to simulate the behavior of processing and rendering values."""
        arg3: float = 7.0
        arg4: AtRender[float] = 8.0
        return arg1 + arg2 + arg3 + arg4 + int(ARG1) + int(ARG2)


def fn(arg5: int, arg6: AtRender[int]) -> float:
    """A mock function to simulate processing of rendered values with integer inputs."""
    arg7: float = 7.0
    arg8: AtRender[float] = 8.0
    return arg5 + arg6 + arg7 + arg8 + int(ARG1) + int(ARG2)


@workflow()
def to_int_bad(num: int, should_double: bool | Any) -> int | float:
    """This workflow is in-valid because the boolean is NOT Annotated with AtRender and resolved at render time by dewret.

    TODO: understand and document if this is possible in dewret
    """
    return increment(num=num) if should_double else sum(left=num, right=num)


@workflow()
def to_int(num: int, should_double: AtRender[bool]) -> int | float:
    """This workflow is valid because the boolean is Annotated with AtRender, and resolved at render time by dewret.

    The output does not contain a branch on the value of should_double.
    """
    return increment(num=num) if should_double else sum(left=num, right=num)


def test_can_analyze_annotations() -> None:
    """Test that annotations can be correctly analyzed within methods and functions.

    Verifies that the `FunctionAnalyser` finds which arguments
    and global variables are derived from `dewret.annotations.AtRender`.
    """
    my_obj = MyClass()

    analyser = FunctionAnalyser(my_obj.method)
    assert analyser.argument_has("arg1", AtRender, exhaustive=True) is False
    assert analyser.argument_has("arg3", AtRender, exhaustive=True) is False
    assert analyser.argument_has("ARG2", AtRender, exhaustive=True) is False
    assert analyser.argument_has("arg2", AtRender, exhaustive=True) is True
    assert analyser.argument_has("arg2", AtRender, exhaustive=False) is True
    assert (
        analyser.argument_has("arg4", AtRender, exhaustive=True) is False
    )  # Not a global/argument
    assert analyser.argument_has("ARG1", AtRender, exhaustive=True) is True
    assert analyser.argument_has("ARG1", AtRender) is False

    analyser = FunctionAnalyser(fn)
    assert analyser.argument_has("arg5", AtRender, exhaustive=True) is False
    assert analyser.argument_has("arg7", AtRender, exhaustive=True) is False
    assert analyser.argument_has("ARG2", AtRender, exhaustive=True) is False
    assert analyser.argument_has("arg6", AtRender, exhaustive=True) is True
    assert analyser.argument_has("arg6", AtRender, exhaustive=False) is True
    assert (
        analyser.argument_has("arg8", AtRender, exhaustive=True) is False
    )  # Not a global/argument
    assert analyser.argument_has("ARG1", AtRender, exhaustive=True) is True
    assert analyser.argument_has("ARG1", AtRender) is False


def test_at_render_bad() -> None:
    """Test the rendering of workflows with exceptions handling."""
    with pytest.raises(TaskException) as _:
        result = to_int_bad(num=increment(num=3), should_double=True)
        construct(result, simplify_ids=True)


def test_at_render() -> None:
    """Test the rendering of workflows with `dewret.annotations.AtRender`."""
    result = to_int(num=increment(num=3), should_double=True)
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    rendered = subworkflows["__root__"]
    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          increment-1-num:
            default: 3
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: to_int-1/out
            type:
            - int
            - float
        steps:
          increment-1:
            in:
              num:
                source: increment-1-num
            out:
            - out
            run: increment
          to_int-1:
            in:
              num:
                source: increment-1/out
              should_double:
                default: True
            out:
            - out
            run: to_int
    """)

    result = to_int(num=increment(num=3), should_double=False)
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    rendered = subworkflows["__root__"]
    assert rendered == yaml.safe_load("""
        cwlVersion: 1.2
        class: Workflow
        inputs:
          increment-1-num:
            default: 3
            label: num
            type: int
        outputs:
          out:
            label: out
            outputSource: to_int-1/out
            type:
            - int
            - float
        steps:
          increment-1:
            in:
              num:
                source: increment-1-num
            out:
            - out
            run: increment
          to_int-1:
            in:
              num:
                source: increment-1/out
              should_double:
                default: False
            out:
            - out
            run: to_int
    """)


def test_at_render_between_modules() -> None:
    """Test rendering of workflows across different modules using `dewret.annotations.AtRender`."""
    result = try_nothing()
    wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    subworkflows["__root__"]


list_2: Fixed[list[int]] = [0, 1, 2, 3]


def test_can_loop_over_fixed_length() -> None:
    """Test looping over a fixed-length list using `dewret.annotations.Fixed`."""

    @workflow()
    def loop_over_lists(list_1: list[int]) -> list[int]:
        result = []
        for a, b in zip(list_1, list_2, strict=False):
            result.append(a + b + len(list_2))
        return result

    with set_configuration(flatten_all_nested=True):
        result = loop_over_lists(list_1=[5, 6, 7, 8])
        wkflw = construct(result, simplify_ids=True)
    subworkflows = render(wkflw, allow_complex_types=True)
    rendered = subworkflows["__root__"]
    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          out:
            type: float
            label: out
            expression: '[4 + list_1[0] + list_2[0], 4 + list_1[1] + list_2[1], 4 + list_1[2] + list_2[2],
              4 + list_1[3] + list_2[3]]'
            source:
            - list_1
            - list_2
        steps: {}
    """)



"""Test CLI render function."""

import yaml
import pytest
from pathlib import Path
from dewret.cli import render


def test_cli_call() -> None:
    """Test CLI render function."""
    # Setup
    test_dir = Path(__file__).parent
    repo_root = test_dir.parent

    example_file = repo_root / "example" / "workflow_qs.py"
    output_file = (test_dir / "tmp_file").resolve()

    try:
        # not sure why it is raising an exception
        with pytest.raises(SystemExit) as exc_info:
            render(
                [
                    "--pretty",
                    "--backend",
                    "DASK",
                    "--output",
                    str(output_file),
                    example_file.as_posix(),
                    "increment",
                    "num:3",
                ]
            )

        # Check that it was a successful exit (code 0)
        assert exc_info.value.code == 0

        # Now check the output file
        with open(output_file, "r") as file:
            res = yaml.safe_load(file)

        assert list(res.keys()) == ["class", "cwlVersion", "inputs", "outputs", "steps"]
    finally:
        # Delete the file in the finally block to ensure it's always cleaned up
        if output_file.exists():
            output_file.unlink()



# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CLI for dewret.

Simple CLI for running a workflow (dynamic->static) conversion. It is
more likely to use this tool programmatically, but for CI and toy examples,
this may be of use.
"""

import importlib
import importlib.util
from pathlib import Path
from contextlib import contextmanager
import sys
import re
import yaml
from typing import Any, IO, Generator
from types import ModuleType
import click
import json

from .core import (
    set_configuration,
    set_render_configuration,
    RawRenderModule,
    StructuredRenderModule,
)
from .utils import load_module_or_package
from .render import get_render_method, write_rendered_output

from .tasks import Backend, construct


@click.command()
@click.option(
    "--pretty",
    is_flag=True,
    show_default=True,
    default=False,
    help="Pretty-print output where possible.",
)
@click.option(
    "--backend",
    type=click.Choice(list(Backend.__members__)),
    show_default=True,
    default=Backend.DASK.name,
    help="Backend to use for workflow evaluation.",
)
@click.option("--construct-args", default="simplify_ids:true")
@click.option("--renderer", default="cwl")
@click.option("--renderer-args", default="")
@click.option("--output", default="-")
@click.argument("workflow_py", type=click.Path(exists=True, path_type=Path))
@click.argument("task")
@click.argument("arguments", nargs=-1)
def render(
    workflow_py: Path,
    task: str,
    arguments: list[str],
    pretty: bool,
    backend: Backend,
    construct_args: str,
    renderer: str,
    renderer_args: str,
    output: str,
) -> None:
    """Render a workflow.

    WORKFLOW_PY is the Python file containing workflow.
    TASK is the name of (decorated) task in workflow module.
    ARGUMENTS is zero or more pairs representing constant arguments to pass to the task, in the format `key:val` where val is a JSON basic type.
    """
    sys.path.append(str(workflow_py.parent))
    kwargs = {}
    for arg in arguments:
        if ":" not in arg:
            raise RuntimeError(
                "Arguments should be specified as key:val, where val is a JSON representation of the argument"
            )
        key, val = arg.split(":", 1)
        kwargs[key] = json.loads(val)

    render_module: Path | ModuleType
    if mtch := re.match(r"^([a-z_0-9-.]+)$", renderer):
        try:
            render_module = importlib.import_module(mtch.group(1))
        except ModuleNotFoundError:
            try:
                render_module = importlib.import_module(
                    f"dewret.renderers.{mtch.group(1)}"
                )
            except ModuleNotFoundError as err:
                raise ModuleNotFoundError(
                    f"Unable to find render module {mtch.group(1)} in PYTHONPATH or dewret.renderers."
                ) from err
        if not isinstance(render_module, RawRenderModule) and not isinstance(
            render_module, StructuredRenderModule
        ):
            raise NotImplementedError(
                "The imported render module does not seem to match the `RawRenderModule` or `StructuredRenderModule` protocols."
            )
    elif renderer.startswith("@"):
        render_module = Path(renderer[1:])
    else:
        raise RuntimeError(
            "Renderer argument should be a known dewret renderer, or '@FILENAME' where FILENAME is a renderer"
        )

    if construct_args.startswith("@"):
        with Path(construct_args[1:]).open() as construct_args_f:
            construct_kwargs = yaml.safe_load(construct_args_f)
    elif not construct_args:
        construct_kwargs = {}
    else:
        construct_kwargs = dict(pair.split(":") for pair in construct_args.split(","))

    renderer_kwargs: dict[str, Any]
    if renderer_args.startswith("@"):
        with Path(renderer_args[1:]).open() as renderer_args_f:
            renderer_kwargs = yaml.safe_load(renderer_args_f)
    elif not renderer_args:
        renderer_kwargs = {}
    else:
        renderer_kwargs = dict(pair.split(":") for pair in renderer_args.split(","))

    if output == "-":

        @contextmanager
        def _opener(key: str, mode: str) -> Generator[IO[Any], None, None]:
            print(" ------ ", key, " ------ ")
            yield sys.stdout
            print()

        opener = _opener
    else:

        @contextmanager
        def _opener(key: str, mode: str) -> Generator[IO[Any], None, None]:
            output_file = output.replace("%", key)
            with Path(output_file).open(mode) as output_f:
                yield output_f

        opener = _opener

    render = get_render_method(render_module, pretty=pretty)
    pkg = "__workflow__"
    workflow = load_module_or_package(pkg, workflow_py)
    task_fn = getattr(workflow, task)

    try:
        with (
            set_configuration(**construct_kwargs),
            set_render_configuration(renderer_kwargs),
        ):
            rendered = render(
                construct(task_fn(**kwargs), **construct_kwargs), **renderer_kwargs
            )
    except Exception as exc:
        import traceback

        print(exc, exc.__cause__, exc.__context__)
        traceback.print_exc()
    else:
        write_rendered_output(rendered, output, opener)



"""Testing example renderer.

'Friendly render', outputting human-readable descriptions.
"""

from textwrap import indent
from typing import Unpack, TypedDict
from dataclasses import dataclass

from dewret.core import set_render_configuration
from dewret.workflow import Workflow, Step, NestedStep
from dewret.render import base_render

from .extra import JUMP


class FrenderRendererConfiguration(TypedDict):
    allow_complex_types: bool


def default_config() -> FrenderRendererConfiguration:
    return FrenderRendererConfiguration({"allow_complex_types": True})


@dataclass
class NestedStepDefinition:
    name: str
    subworkflow_name: str

    @classmethod
    def from_nested_step(cls, nested_step: NestedStep) -> "NestedStepDefinition":
        return cls(name=nested_step.name, subworkflow_name=nested_step.subworkflow.name)

    def render(self) -> str:
        return f"""
A portal called {self.name} to another workflow,
whose name is {self.subworkflow_name}
"""


@dataclass
class StepDefinition:
    name: str

    @classmethod
    def from_step(cls, step: Step) -> "StepDefinition":
        return cls(name=step.name)

    def render(self) -> str:
        return f"""
Something called {self.name}
"""


@dataclass
class WorkflowDefinition:
    name: str
    steps: list[StepDefinition | NestedStepDefinition]

    @classmethod
    def from_workflow(cls, workflow: Workflow) -> "WorkflowDefinition":
        steps: list[StepDefinition | NestedStepDefinition] = []
        for step in workflow.indexed_steps.values():
            if isinstance(step, Step):
                steps.append(StepDefinition.from_step(step))
            elif isinstance(step, NestedStep):
                steps.append(NestedStepDefinition.from_nested_step(step))
            else:
                raise RuntimeError(f"Unrecognised step type: {type(step)}")

        try:
            name = workflow.name
        except NameError:
            name = "Work Doe"
        return cls(name=name, steps=steps)

    def render(self) -> str:
        steps = "\n".join("* " + indent(step.render(), "  ")[3:] for step in self.steps)
        return f"""
I found a workflow called {self.name}.
It has {len(self.steps)} steps!
They are:
{steps}
It probably got made with JUMP={JUMP}
"""


def render_raw(
    workflow: Workflow, **kwargs: Unpack[FrenderRendererConfiguration]
) -> dict[str, str]:
    """Render to a dict-like structure.

    Args:
        workflow: workflow to evaluate result.
        **kwargs: additional configuration arguments - these should match CWLRendererConfiguration.

    Returns:
        Reduced form as a native Python dict structure for
        serialization.
    """
    # TODO: work out how to handle these hints correctly.
    set_render_configuration(kwargs)  # type: ignore
    return base_render(
        workflow, lambda workflow: WorkflowDefinition.from_workflow(workflow).render()
    )



from portray import config, render  # type: ignore
from typing import Any
import yaml

# Note that this does not work for reloads.
mkdocs = config.mkdocs


def _mkdocs(directory: str, **overrides) -> dict:  # type: ignore
    superfences = yaml.unsafe_load("""
      preserve_tabs: true
      custom_fences:
        # Mermaid diagrams
        - name: mermaid
          class: mermaid
          format: !!python/name:pymdownx.superfences.fence_code_format
    """)
    overrides.setdefault("markdown_extensions", [])
    for n, ext in enumerate(overrides["markdown_extensions"]):
        if ext == "pymdownx.superfences":
            ext = {"pymdownx.superfences": {}}
        elif isinstance(ext, dict) and len(ext) == 1 and "pymdownx.superfences" in ext:
            ...
        else:
            continue
        ext["pymdownx.superfences"].update(superfences)
        overrides["markdown_extensions"][n] = ext
    res: dict[Any, Any] = mkdocs(directory, **overrides)
    return res


config.mkdocs = _mkdocs

mkdocs_render = render.mkdocs


def _mkdocs_render(config: dict[str, Any]) -> Any:
    """Ensures the original config can be reused."""
    original_config = list(config.get("markdown_extensions", []))
    if original_config and "pymdownx.superfences" in original_config[0]:
        original_config = original_config[0]["pymdownx.superfences"]
        result = mkdocs_render(config)
        config["markdown_extensions"][0]["pymdownx.superfences"] = original_config
    else:
        result = mkdocs_render(config)
    return result


render.mkdocs = _mkdocs_render

if __name__ == "__main__":
    from portray.cli import cli  # type: ignore

    cli()



"""Check complex nested structures and expressions can mix."""

import yaml
import math
from dewret.workflow import param
from dewret.tasks import construct
from dewret.renderers.cwl import render

from ._lib.extra import reverse_list, max_list


def test_can_supply_nested_raw() -> None:
    """TODO: The structures are important for future CWL rendering."""
    pi = param("pi", math.pi)
    result = reverse_list(to_sort=[1.0, 3.0, pi])
    workflow = construct(max_list(lst=result + result), simplify_ids=True)
    # assert workflow.find_parameters() == {
    #    pi
    # }

    # NB: This is not currently usefully renderable in CWL.
    # However, the structures are important for future CWL rendering.

    rendered = render(workflow)["__root__"]
    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          pi:
            default: 3.141592653589793
            label: pi
            type: float
        outputs:
          out:
            label: out
            outputSource: max_list-1/out
            type:
            - int
            - float
        steps:
          max_list-1:
            in:
              lst:
                source: reverse_list-1/out
                valueFrom: $(2*self)
            out:
            - out
            run: max_list
          reverse_list-1:
            in:
              to_sort:
                valueFrom: $((1.0, 3.0, inputs.pi))
            out:
            - out
            run: reverse_list
    """)



# Copyright 2014 Flax & Teal Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dask backend.

Lazy-evaluation via `dask.delayed`.
"""

from dask.delayed import delayed, DelayedLeaf
from dask.config import config
from typing import Protocol, runtime_checkable, Any, cast
from concurrent.futures import ThreadPoolExecutor
from dewret.workflow import Workflow, Lazy, StepReference, Target


@runtime_checkable
class Delayed(Protocol):
    """Description of a dask `delayed`.

    Since `dask.delayed` does not have a hintable type, this
    stands in its place, making sure that all the features of a
    `dask.delayed` are available.

    More info: https://github.com/dask/dask/issues/7779
    """

    @property
    def __dask_graph__(self):  # type: ignore
        """Retrieve the dask graph."""
        ...

    def compute(self, __workflow__: Workflow | None) -> StepReference[Any]:
        """Evaluate this `dask.delayed`.

        Evaluate a delayed (dask lazy-evaluated) function. dewret
        will have replaced it with a wrapper that expects a `Workflow`
        and all arguments will already be known to the wrapped `delayed`
        so the signature here is simple.

        Args:
            __workflow__: `Workflow` that this is tied to, if applicable.

        Returns:
            Reference to the final output step.
        """
        ...


def unwrap(task: Lazy) -> Target:
    """Unwraps a lazy-evaluated function to get the function.

    In recent dask (>=2024.3) this works with inspect.wraps, but earlier
    versions do not have the `__wrapped__` property.

    Args:
        task: task to be unwrapped.

    Returns:
        Original target.

    Raises:
        RuntimeError: if the task is not a wrapped function.
    """
    if not isinstance(task, DelayedLeaf):
        raise RuntimeError("Task is not for this backend")
    if not callable(task):
        raise RuntimeError("Task is not actually a callable")
    return cast(Target, task._obj)


def is_lazy(task: Any) -> bool:
    """Checks if a task is really a lazy-evaluated function for this backend.

    Args:
        task: suspected lazy-evaluated function.

    Returns:
        True if so, False otherwise.
    """
    return isinstance(task, Delayed) or (
        isinstance(task, tuple | list) and all(is_lazy(elt) for elt in task)
    )


lazy = delayed


def run(
    workflow: Workflow | None,
    task: Lazy | list[Lazy] | tuple[Lazy],
    thread_pool: ThreadPoolExecutor | None = None,
    **kwargs: Any,
) -> Any:
    """Execute a task as the output of a workflow.

    Runs a task with dask.

    Args:
        workflow: `Workflow` in which to record the execution.
        task: `dask.delayed` function, wrapped by dewret, that we wish to compute.
        thread_pool: custom thread pool for executing workflows, copies in correct values for contextvars to each thread before they are accessed by a dask worker.
        **kwargs: any configuration arguments for this backend.
    """
    # def _check_delayed was here, but we decided to delegate this to dask

    if isinstance(task, Delayed) and is_lazy(task):
        computable = task
    else:
        computable = delayed(task)
    config["pool"] = thread_pool
    result = computable.compute(__workflow__=workflow)
    return result



"""Verify CWL can be made with split up and nested calls."""

import yaml
from attr import define
from dataclasses import dataclass
from typing import Iterable
from dewret.tasks import task, construct, workflow
from dewret.core import set_configuration
from dewret.renderers.cwl import render

STARTING_NUMBER: int = 23


@define
class SplitResult:
    """Test class showing two named values, using attrs."""

    first: int
    second: float


@dataclass
class SplitResultDataclass:
    """Test class showing two named values, using dataclasses."""

    first: int
    second: float


@task()
def combine(left: int, right: float) -> float:
    """Sum two values."""
    return left + right


@task()
def list_cast(iterable: Iterable[float]) -> list[float]:
    """Converts an iterable structure with float elements into a list of floats."""
    return list(iterable)


@task()
def pair(left: int, right: float) -> tuple[int, float]:
    """Pairs two values."""
    return (left, right)


@workflow()
def algorithm() -> float:
    """Sum two split values."""
    return combine(left=split().first, right=split().second)


@workflow()
def algorithm_with_pair() -> tuple[int, float]:
    """Pairs two split dataclass values."""
    return pair(left=split_into_dataclass().first, right=split_into_dataclass().second)


@workflow()
def algorithm_with_dataclasses() -> float:
    """Sums two split dataclass values."""
    return combine(
        left=split_into_dataclass().first, right=split_into_dataclass().second
    )


@task()
def split() -> SplitResult:
    """Create a split result with two fields."""
    return SplitResult(first=1, second=2)


@task()
def split_into_dataclass() -> SplitResultDataclass:
    """Create a result with two fields."""
    return SplitResultDataclass(first=1, second=2)


def test_subworkflow() -> None:
    """Check whether we can link between multiple steps and have parameters.

    Produces CWL that has references between multiple steps.
    """
    workflow = construct(split(), simplify_ids=True)
    rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          out:
            label: out
            outputSource: split-1/out
            type: record
            fields:
                first:
                    label: first
                    type: int
                second:
                    label: second
                    type: float
        steps:
          split-1:
            in: {}
            out:
              first:
                label: first
                type: int
              second:
                label: second
                type: float
            run: split
    """)


def test_field_of_subworkflow() -> None:
    """Tests whether a directly-output nested task can have fields."""
    workflow = construct(split().first, simplify_ids=True)
    rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          first:
            label: first
            outputSource: split-1/first
            type: int
        steps:
          split-1:
            in: {}
            out:
              first:
                label: first
                type: int
              second:
                label: second
                type: float
            run: split
    """)


def test_field_of_subworkflow_into_dataclasses() -> None:
    """Tests whether a directly-output nested task can have fields."""
    workflow = construct(split_into_dataclass().first, simplify_ids=True)
    rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          first:
            label: first
            outputSource: split_into_dataclass-1/first
            type: int
        steps:
          split_into_dataclass-1:
            in: {}
            out:
              first:
                label: first
                type: int
              second:
                label: second
                type: float
            run: split_into_dataclass
    """)


def test_complex_field_of_subworkflow() -> None:
    """Tests whether a task can sum complex structures."""
    with set_configuration(flatten_all_nested=True):
        workflow = construct(algorithm(), simplify_ids=True)
        rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          out:
            label: out
            outputSource: combine-1/out
            type: float
        steps:
          combine-1:
            in:
                left:
                    source: split-1/first
                right:
                    source: split-1/second
            out: [out]
            run: combine
          split-1:
            in: {}
            out:
              first:
                label: first
                type: int
              second:
                label: second
                type: float
            run: split
    """)


def test_complex_field_of_subworkflow_with_dataclasses() -> None:
    """Tests whether a task can insert result fields into other steps."""
    with set_configuration(flatten_all_nested=True):
        result = algorithm_with_dataclasses()
        workflow = construct(result, simplify_ids=True)
        rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          out:
            label: out
            outputSource: combine-1/out
            type: float
        steps:
          combine-1:
            in:
                left:
                    source: split_into_dataclass-1/first
                right:
                    source: split_into_dataclass-1/second
            out: [out]
            run: combine
          split_into_dataclass-1:
            in: {}
            out:
              first:
                label: first
                type: int
              second:
                label: second
                type: float
            run: split_into_dataclass
    """)


def test_pair_can_be_returned_from_step() -> None:
    """Tests whether a task can insert result fields into other steps."""
    with set_configuration(flatten_all_nested=True):
        workflow = construct(algorithm_with_pair(), simplify_ids=True)
        rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          out:
            label: out
            outputSource: pair-1/out
            items:
              - int
              - float
            type: array
        steps:
          pair-1:
            in:
                left:
                    source: split_into_dataclass-1/first
                right:
                    source: split_into_dataclass-1/second
            out: [out]
            run: pair
          split_into_dataclass-1:
            in: {}
            out:
              first:
                label: first
                type: int
              second:
                label: second
                type: float 
            run: split_into_dataclass
    """)


def test_list_can_be_returned_from_step() -> None:
    """Tests whether a task can insert result fields into other steps."""
    with set_configuration(flatten_all_nested=True):
        workflow = construct(
            list_cast(iterable=algorithm_with_pair()), simplify_ids=True
        )
        rendered = render(workflow)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          out:
            label: out
            outputSource: list_cast-1/out
            items: float
            type: array
        steps:
          list_cast-1:
            in:
                iterable:
                    source: pair-1/out
            out: [out]
            run: list_cast
          pair-1:
            in:
                left:
                    source: split_into_dataclass-1/first
                right:
                    source: split_into_dataclass-1/second
            out: [out]
            run: pair
          split_into_dataclass-1:
            in: {}
            out:
              first:
                label: first
                type: int
              second:
                label: second
                type: float
            run: split_into_dataclass
    """)



"""Check field management works."""

from __future__ import annotations
import yaml
from dataclasses import dataclass

from typing import Unpack, TypedDict

from dewret.tasks import task, construct, workflow
from dewret.core import set_configuration
from dewret.workflow import param, StepReference
from dewret.renderers.cwl import render
from dewret.annotations import Fixed

from ._lib.extra import mod10, sum, pi


@dataclass
class Sides:
    """A dataclass representing the sides with `left` and `right` integers."""

    left: int
    right: int


SIDES: Sides = Sides(3, 6)


@workflow()
def sum_sides() -> float:
    """Workflow that returns the sum of the `left` and `right` sides."""
    return sum(left=SIDES.left, right=SIDES.right)


def test_fields_of_parameters_usable() -> None:
    """Test that fields of parameters can be used to construct and render a workflow correctly."""
    result = sum_sides()
    wkflw = construct(result, simplify_ids=True)
    rendered = render(wkflw, allow_complex_types=True)["sum_sides-1"]

    assert rendered == yaml.safe_load("""
      class: Workflow
      cwlVersion: 1.2
      inputs:
        SIDES:
          label: SIDES
          default:
            left: 3
            right: 6
          type: record
          fields:
            left:
              default: 3
              label: left
              type: int
            right:
              default: 6
              label: right
              type: int
          label: SIDES
      outputs:
        out:
          label: out
          outputSource: sum-1-1/out
          type:
          - int
          - float
      steps:
        sum-1-1:
          in:
            left:
              source: SIDES/left
            right:
              source: SIDES/right
          out:
          - out
          run: sum
    """)


@dataclass
class MyDataclass:
    """A dataclass with nested references to itself, containing `left` and `right` fields."""

    left: int
    right: "MyDataclass"


def test_can_get_field_reference_from_parameter() -> None:
    """Test that field references can be retrieved from a parameter when constructing a workflow."""
    my_param = param("my_param", typ=MyDataclass)
    result = sum(
        left=my_param.left, right=sum(left=my_param.right.left, right=my_param.left)
    )
    wkflw = construct(result, simplify_ids=True)
    params = {(str(p), p.__type__) for p in wkflw.find_parameters()}

    assert params == {("my_param", MyDataclass)}
    rendered = render(wkflw, allow_complex_types=True)["__root__"]
    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          my_param:
            label: my_param
            type: MyDataclass
        outputs:
          out:
            label: out
            outputSource: sum-1/out
            type:
            - int
            - float
        steps:
          sum-1:
            in:
              left:
                source: my_param/left
              right:
                source: sum-2/out
            out:
            - out
            run: sum
          sum-2:
            in:
              left:
                source: my_param/right/left
              right:
                source: my_param/left
            out:
            - out
            run: sum
    """)


def test_can_get_field_reference_if_parent_type_has_field() -> None:
    """Test that a field reference is retrievable if the parent type has that field."""

    @dataclass
    class MyDataclass:
        left: int

    my_param = param("my_param", typ=MyDataclass)
    result = sum(left=my_param.left, right=my_param.left)
    wkflw = construct(result, simplify_ids=True)
    param_reference = list(wkflw.find_parameters())[0]

    assert str(param_reference.left) == "my_param/left"
    assert param_reference.left.__type__ == int


def test_can_get_go_upwards_from_a_field_reference() -> None:
    """Test that it's possible to move upwards in the hierarchy from a field reference."""

    @dataclass
    class MyDataclass:
        left: int
        right: "MyDataclass"

    my_param = param("my_param", typ=MyDataclass)
    result = sum(left=my_param.left, right=my_param.left)
    construct(result, simplify_ids=True)

    back = my_param.right.left.__field_up__()  # type: ignore
    assert str(back) == "my_param/right"
    assert back.__type__ == MyDataclass


def test_can_get_field_references_from_dataclass() -> None:
    """Test that field references can be extracted from a dataclass and used in workflows."""

    @dataclass
    class MyDataclass:
        left: int
        right: float

    @workflow()
    def test_dataclass(my_dataclass: MyDataclass) -> MyDataclass:
        result: MyDataclass = MyDataclass(left=mod10(num=my_dataclass.left), right=pi())
        return result

    @workflow()
    def get_left(my_dataclass: MyDataclass) -> int:
        return my_dataclass.left

    result = get_left(
        my_dataclass=test_dataclass(my_dataclass=MyDataclass(left=3, right=4.0))
    )
    wkflw = construct(result, simplify_ids=True)

    assert isinstance(wkflw.result, StepReference)
    assert str(wkflw.result) == "get_left-1"
    assert wkflw.result.__type__ == int


class MyDict(TypedDict):
    """A typed dictionary with `left` as an integer and `right` as a float."""

    left: int
    right: float


def test_can_get_field_references_from_typed_dict() -> None:
    """Test that field references can be extracted from a custom typed dictionary and used in workflows."""

    @workflow()
    def test_dict(**my_dict: Unpack[MyDict]) -> MyDict:
        result: MyDict = {"left": mod10(num=my_dict["left"]), "right": pi()}
        return result

    result = test_dict(left=3, right=4.0)
    wkflw = construct(result, simplify_ids=True)

    assert isinstance(wkflw.result, StepReference)
    assert str(wkflw.result["left"]) == "test_dict-1/left"
    assert wkflw.result["left"].__type__ == int


@dataclass
class MyListWrapper:
    """A dataclass that wraps a list of integers."""

    my_list: list[int]


def test_can_iterate() -> None:
    """Test iteration over a list of tasks and validate positional argument handling."""

    @task()
    def test_task(alpha: int, beta: float, charlie: bool) -> int:
        """Task that adds `alpha` and `beta` and returns the integer result."""
        return int(alpha + beta)

    @task()
    def test_list() -> list[int | float]:
        """Task that returns a list containing an integer and a float."""
        return [1, 2.0]

    @workflow()
    def test_iterated() -> int:
        """Workflow that tests task iteration over a list."""
        # We ignore the type as mypy cannot confirm that the length and types match the args.
        return test_task(*test_list())  # type: ignore

    with set_configuration(allow_positional_args=True, flatten_all_nested=True):
        result = test_iterated()
        wkflw = construct(result, simplify_ids=True)

    rendered = render(wkflw, allow_complex_types=True)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs: {}
        outputs:
          out:
            label: out
            outputSource: test_task-1/out
            type: int
        steps:
          test_list-1:
            in: {}
            out:
            - out
            run: test_list
          test_task-1:
            in:
              alpha:
                source: test_list-1[0]
              beta:
                source: test_list-1[1]
              charlie:
                source: test_list-1[2]
            out:
            - out
            run: test_task
    """)

    assert isinstance(wkflw.result, StepReference)
    assert wkflw.result._.step.positional_args == {
        "alpha": True,
        "beta": True,
        "charlie": True,
    }

    @task()
    def test_list_2() -> MyListWrapper:
        """Task that returns a `MyListWrapper` containing a list of integers."""
        return MyListWrapper(my_list=[1, 2])

    @workflow()
    def test_iterated_2(my_wrapper: MyListWrapper) -> int:
        """Workflow that tests iteration over a list in a `MyListWrapper`."""
        # mypy cannot confirm argument types match.
        return test_task(*my_wrapper.my_list)  # type: ignore

    with set_configuration(allow_positional_args=True, flatten_all_nested=True):
        result = test_iterated_2(my_wrapper=test_list_2())
        wkflw = construct(result, simplify_ids=True)

    @task()
    def test_list_3() -> Fixed[list[tuple[int, int]]]:
        """Task that returns a list of integer tuples wrapped in a `dewret.annotations.Fixed` type."""
        return [(0, 1), (2, 3)]

    @workflow()
    def test_iterated_3(param: Fixed[list[tuple[int, int]]]) -> int:
        """Workflow that iterates over a list of integer tuples and performs operations."""
        # mypy cannot confirm argument types match.
        retval = mod10(*test_list_3()[0])  # type: ignore
        for pair in param:
            a, b = pair
            retval += a + b
        return mod10(retval)

    with set_configuration(allow_positional_args=True, flatten_all_nested=True):
        result = test_iterated_3(param=[(0, 1), (2, 3)])
        wkflw = construct(result, simplify_ids=True)

    rendered = render(wkflw, allow_complex_types=True)["__root__"]

    assert rendered == yaml.safe_load("""
        class: Workflow
        cwlVersion: 1.2
        inputs:
          param:
            default:
            - - 0
              - 1
            - - 2
              - 3
            items: array
            label: param
            type: array
        outputs:
          out:
            label: out
            outputSource: mod10-2/out
            type: int
        steps:
          mod10-1:
            in:
              num:
                source: test_list_3-1[0][0]
            out:
            - out
            run: mod10
          mod10-2:
            in:
              num:
                valueFrom: $(inputs.param[0][0] + inputs.param[0][1] + inputs.param[1][0] + inputs.param[1][1] + self)
                source: mod10-1/out
            out:
            - out
            run: mod10
          test_list_3-1:
            in: {}
            out:
            - out
            run: test_list_3
    """)


def test_can_use_plain_dict_fields() -> None:
    """Test the use of plain dictionary fields in workflows."""

    @workflow()
    def test_dict(left: int, right: float) -> dict[str, float | int]:
        result: dict[str, float | int] = {"left": mod10(num=left), "right": pi()}
        return result

    with set_configuration(allow_plain_dict_fields=True):
        result = test_dict(left=3, right=4.0)
        wkflw = construct(result, simplify_ids=True)
        assert isinstance(wkflw.result, StepReference)
        assert str(wkflw.result["left"]) == "test_dict-1/left"
        assert wkflw.result["left"].__type__ == int | float


@dataclass
class IndexTest:
    """A dataclass for testing indexed fields, containing a `left` field that is a list of integers."""

    left: Fixed[list[int]]


def test_can_configure_field_separator() -> None:
    """Test the ability to configure the field separator in workflows."""

    @task()
    def test_sep() -> IndexTest:
        return IndexTest(left=[3])

    with set_configuration(field_index_types="int"):
        result = test_sep().left[0]
        wkflw = construct(result, simplify_ids=True)
        render(wkflw, allow_complex_types=True)["__root__"]
        assert str(wkflw.result) == "test_sep-1/left[0]"

    with set_configuration(field_index_types="int,str"):
        result = test_sep().left[0]
        wkflw = construct(result, simplify_ids=True)
        render(wkflw, allow_complex_types=True)["__root__"]
        assert str(wkflw.result) == "test_sep-1[left][0]"

    with set_configuration(field_index_types=""):
        result = test_sep().left[0]
        wkflw = construct(result, simplify_ids=True)
        render(wkflw, allow_complex_types=True)["__root__"]
        assert str(wkflw.result) == "test_sep-1/left/0"
